import java.util.*;
import java.io.*;
import java.sql.*;




/*-------------------------------------------------------------- LoadGGLE class
|																			  |
|			Class to load the distributed GGLE db from a xml file			  |
|																			  |
|																			  |
|																			  |
| 																			  |
\----------------------------------------------------------------------------*/
public class DistGgleLoad
{
	//--------------------------------------------------------- paramenters
	//CHANGE INPUT FILE HERE											//|
	public static final String FILE = "output.xml";						//|
																		//|
	// CHANGE DRIVER STRING HERE										//|
	public static final String DRIVER =									//|
		"nl.cwi.monetdb.jdbc.MonetDriver";								//|
		//"com.mysql.jdbc.Driver";										//|
																		//|
	// CHANGE DB CONNECTION STRING(S) HERE								//|
																		//|
	public static final String DB1 =									//|
		"jdbc:monetdb://localhost:45123/ggle_1";						//|
		//"jdbc:mysql://localhost/ggle?user=ggle&password=gsa#1";		//|
																		//|
	public static final String DB2 =									//|
		"jdbc:monetdb://localhost:45223/ggle_2";						//|
																		//|
	// SET BATCHSIZE HERE												//|
	public static final int BATCHSIZE = 20;								//|

	// SET MAXBATCHSIZE HERE											//|
	public static final int MAXBATCHSIZE = 1000;						//|
																		//|
	//--------------------------------------------------------------------/



	/*-------------------------------------------------------- showConnection()
	|																		  |
	| 																		  |
	\------------------------------------------------------------------------*/
	public static void showConnection( Connection con ) throws SQLException
	{

		DatabaseMetaData d;
		String dbName, user;
		ResultSet res;

		// -------- retrive db info
		d 	   = con.getMetaData();
		res	   = d.getSchemas();
		dbName = d.getDatabaseProductName() + " " +  d.getDatabaseProductVersion();
		user   = d.getUserName();

		System.out.println("Connected to: " + dbName + " as " + user);
		//System.out.println("Auto commit: " + con.getAutoCommit() );
		System.out.println("Open DB(s):");

		while( res.next() )
		{
			System.out.println( res.getString(2) );
		}
		res.close();
		System.out.println("***********************************\n\n");
	}


	/*------------------------------------------------------------------ main()
	|																		  |
	| main function															  |
	\------------------------------------------------------------------------*/
	public static void main (String argv[])
	{
		// --------------------------------------------- variable declarations

		Connection 			con1, con2	 = null;	// the connection pointers
		ParallelConnection	pcon1, pcon2 = null;
		ResultSet res;
		Entry entry;
		int stat = 0;				 // status report from nextEntry()
		long start = 0, end = 0;
		int EntryId = 0, LinkId = 1;	   // uid for Entry/Link Table
		String sql = "";



		// --------------------------------------------- Open the db connection
		try
		{
			// get the driver
			Class.forName(DRIVER);

			// open connecvtions
			con1 = DriverManager.getConnection(DB1, "monetdb", "monetdb");
			con2 = DriverManager.getConnection(DB2, "monetdb", "monetdb");

			// turn off auto commit
			con1.setAutoCommit( false );
			con2.setAutoCommit( false );

			// show connection details
			showConnection(con1);
			showConnection(con2);

		}
		// --- connection(s) could not be opened -> exit
		catch (Exception e)
		{
			System.out.println(e);
			return;
		}

		// -------------------------- create a new xml parser on the given file
		XMLparser xml = new XMLparser(FILE);

		// ---- file couldn't be opned -> exit
		if( xml == null )
			return;




		System.out.println("Loading data ... ");
		pcon1 = null; pcon2 = null;
		//clearDB();


		// ----------------------------- create and start pconnection threads
		pcon1 = new ParallelConnection( con1 );
		pcon2 = new ParallelConnection( con2 );

		pcon1.start();
		pcon2.start();



		// ------------------- create all statements and temp table
		try
		{
			pcon1.createStatement( "stmt" );
			pcon2.createStatement( "stmt" );


			sql = "CREATE TABLE Tmp " +
				"(	 EntryId	INT	UNIQUE 	 NOT NULL, " +
				"	 URL		VARCHAR(200) NOT NULL )";

			pcon1.executeUpdate( "stmt", sql );
			pcon2.executeUpdate( "stmt", sql );


			sql = "INSERT INTO Tmp VALUES (?, ?)";

			pcon1.prepareStatement(	"temp", sql  );
			pcon2.prepareStatement(	"temp", sql  );


			sql = "INSERT INTO Entry VALUES (?, ?, ?, ?, ?, ?, ?, ?)";

			pcon1.prepareStatement(	"entry", sql  );
			pcon2.prepareStatement(	"entry", sql  );


			sql = "INSERT INTO Link(LinkId, Entry, domain, path, page, url) "
				+ "  VALUES (?, ?, ?, ?, ?, ?)";

			pcon1.prepareStatement("link", sql	);
			pcon2.prepareStatement("link", sql	);

			sql = "UPDATE Link SET LinksTo = ? WHERE LinkId = ? ";

			pcon1.prepareStatement("update", sql	);
			pcon2.prepareStatement("update", sql	);



		}
		// --- statements or temp table could not be created -> exit
		catch (Exception e)
		{
			System.out.println(e);
			return;
		}


		// ------------------------------------------- loop through all entries
		start = System.currentTimeMillis();
		while( stat != -1 )
		{

			// -------------------------------------- parse the next entry
			try
			{
				stat = xml.nextEntry();
				EntryId++;
			}
			catch(IOException e)
			{
				System.out.println(
					"ERROR READING FROM FILE!!\n IO EXCEPTION: " + e );
				stat = -1;
				break;
			}
			// --- error reading next entry -> skip
			if ( stat <= 0 )
				continue;

			// ------------------------------------------------- get entry
			entry = xml.getEntry();
			entry.setEntryId(EntryId);

			// ---------------------------------- send entry to correct db
			// ------------------------ even entryId -> db1
			if( (EntryId % 2) == 0 )
			{
				LinkId = entry.addBatch
				(
					pcon1.getPStatement("entry"),
					pcon1.getPStatement("link"),
					LinkId
				);

				// --- insert a copy into the other db's temp table
				try
				{
					pcon2.setInt	( "temp", 1, EntryId 		);
					pcon2.setString	( "temp", 2, entry.getUrl()	);
					pcon2.addBatch	( "temp");
				}
				catch( Exception e )
				{
					System.out.println( e );
				}
			}
			// ------------------------- odd entryId -> db2
			else
			{
				LinkId = entry.addBatch
				(
					pcon2.getPStatement("entry"),
					pcon2.getPStatement("link"),
					LinkId
				);

				// --- insert a copy into the other db's temp table
				try
				{
					pcon1.setInt	( "temp", 1, EntryId		 );
					pcon1.setString	( "temp", 2, entry.getUrl()	 );
					pcon1.addBatch	( "temp");
				}
				catch( Exception e )
				{
					System.out.println( e );
				}
			}

			// -------------------- if batch size is reached execute batch
			if( (EntryId % BATCHSIZE) == 0 )
			{
				pcon1.executeBatch("entry"); pcon2.executeBatch("entry");

				pcon1.executeBatch("link");  pcon2.executeBatch("link");

				pcon1.executeBatch("temp");  pcon2.executeBatch("temp");
			}

		} // --- end entry loop


		// ------------------------------------------- execute final batch
		pcon1.executeBatch("entry"); pcon2.executeBatch("entry");

		pcon1.executeBatch("link");  pcon2.executeBatch("link");

		pcon1.executeBatch("temp");  pcon2.executeBatch("temp");

		/*pcon1.executeUpdate("stmt", "DROP TABLE Temp" );
		pcon2.executeUpdate("stmt", "DROP TABLE Temp" );*/


		// ------------------------------------------------------- update links
		System.out.println("Updating links ...."); System.out.flush();



		sql = "SELECT EntryId, LinkId FROM Entry E, Link L " +
			  "		WHERE E.URL = L.URL " +
			  "UNION " +
			  "SELECT EntryId, LinkId FROM Tmp E, Link L " +
			  "		WHERE E.URL = L.URL";


		pcon1.executeQuery ( "stmt", sql	);
		pcon2.executeQuery ( "stmt", sql	);

		res = pcon1.getResult("stmt");

		try
		{
			int count = 0;

			while( res.next() )
			{
				pcon1.setInt	( "update", 1, res.getInt( 1 ) );
				pcon1.setInt	( "update", 2, res.getInt( 2 ) );
				pcon1.addBatch	( "update" );

				if( (count % MAXBATCHSIZE ) == 0 )
					pcon1.executeBatch( "update" );

				count++;

			}
			pcon1.executeBatch( "update" );

			res = pcon2.getResult("stmt");

			count = 0;

			while( res.next() )
			{
				pcon2.setInt	( "update", 1, res.getInt( 1 ) );
				pcon2.setInt	( "update", 2, res.getInt( 2 ) );
				pcon2.addBatch	( "update" );

				if( (count % MAXBATCHSIZE ) == 0 )
					pcon2.executeBatch( "update" );

				count++;
			}
			pcon2.executeBatch( "update" );

			pcon1.executeUpdate( "stmt", "DROP TABLE Tmp");
			pcon2.executeUpdate( "stmt", "DROP TABLE Tmp");


			pcon1.halt();
			pcon2.halt();

			// ---------------------------------- clean up and close connection
			con1.commit(); con2.commit();
			end = System.currentTimeMillis();

			xml.cleanUp();
			con1.close();
			con2.close();


		    /*System.out.println("Thread1 terminated: " + !pcon1.isAlive() );
		    System.out.println("Thread2 terminated: " + !pcon2.isAlive() );

		    System.out.println("con1 closed: " + con1.isClosed() );
		    System.out.println("con2 closed: " + con2.isClosed() );*/

		}
		catch(Exception e)
		{
			System.out.println(e);
		}

		int min, sec, ms;

		Calendar cal = Calendar.getInstance();

		cal.setTime( new java.util.Date(end-start) );
		min = cal.get(Calendar.MINUTE);
		sec = cal.get(Calendar.SECOND);
		ms	= cal.get(Calendar.MILLISECOND);

		System.out.println(
			"Finished loading DB. Loaded " + (EntryId - 2) + " entries, " +
			(LinkId - 1) + " links in " + min + "min " + sec + "sec " + ms + "ms");
	}
}